package core

import (
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/trie"
	"github.com/gogo/protobuf/proto"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
)

type StateDB struct {
	db   *trie.Database
	trie *trie.Trie
}

func NewStateDB(db *trie.Database, root common.Hash) (statedb *StateDB, err error) {
	t, err := trie.New(common.Hash{}, root, db)
	if err != nil {
		return nil, err
	}
	return &StateDB{
		db:   db,
		trie: t,
	}, nil
}

func (s *StateDB) GetMinerObject(miner common.Address) (object *witnessv1.MinerObject) {
	object = &witnessv1.MinerObject{}
	k := crypto.Keccak256(miner.Bytes())
	v, err := s.trie.TryGet(k)
	if err != nil {
		return nil
	}
	err = proto.Unmarshal(v, object)
	if err != nil {
		return nil
	}
	return
}

func (s *StateDB) UpdateMinerObject(miner common.Address, object *witnessv1.MinerObject) (err error) {
	k := crypto.Keccak256(miner.Bytes())
	v, err := proto.Marshal(object)
	if err != nil {
		return
	}

	err = s.trie.TryUpdate(k, v)
	return
}

func (s *StateDB) Commit() (root common.Hash, err error) {
	root, nodes, err := s.trie.Commit(false)
	if err != nil {
		return
	}
	nodesSet := trie.NewMergedNodeSet()
	if nodes != nil {
		err = nodesSet.Merge(nodes)
		if err != nil {
			return
		}
	}
	err = s.db.Update(nodesSet)
	if err != nil {
		return
	}

	err = s.db.Commit(root, true, nil)
	return root, err
}

func (s *StateDB) IterAllObject() (objects []*witnessv1.MinerObject) {
	iter := s.trie.NodeIterator(nil)
	for iter.Next(true) {
		if !iter.Leaf() {
			continue
		}
		v := iter.LeafBlob()
		object := &witnessv1.MinerObject{}
		err := proto.Unmarshal(v, object)
		if err != nil {
			continue
		}
		log.WithField("workload", object.Workload).Debug("miner object")
		objects = append(objects, object)
	}
	return
}
